import sys
import random
import signal
import argparse
from functools import partial, reduce
from itertools import chain

import blessed

from ..grid import Direction, Actions
from .grid import Grid
from .tile import Tile

up, left = ('w', 'k', 'KEY_UP'), ('a', 'h', 'KEY_LEFT')
down, right = ('s', 'j', 'KEY_DOWN'), ('d', 'l', 'KEY_RIGHT')

grid_moves = {}
for keys, direction in zip((up, left, down, right), Direction):
    grid_moves.update(dict.fromkeys(keys, direction))


def grid_dimension(string):
    rows, _, cols = string.partition('x')
    try:
        return int(rows), int(cols)
    except ValueError:
        raise argparse.ArgumentTypeError(
            "grid dimension should look like: '4x4'")


parser = argparse.ArgumentParser(
    description="A game with the objective of merging tiles by moving them.",
    epilog="Use the arrow, wasd or hjkl keys to move the tiles.")
parser.add_argument('grid_dims', metavar='GRID_DIMENSIONS', default=[(4, 4)],
                    type=grid_dimension, nargs='*',
                    help="Dimensions used for grid(s), default: '4x4'")
parser.add_argument('-b', '--base', metavar='N', dest='base', type=int,
                    default=2, help="base value of all tiles")


def draw_score(score, term, end=False):
    msg = "score: " + str(score)
    with term.location(term.width // 2 - len(msg) // 2, 0):
        print(term.bold_on_red(msg) if end else term.bold(msg))


def term_resize(term, grids):
    print(term.clear())

    max_width = (term.width - (len(grids) + 1) * 2) // len(grids)

    for grid in grids:
        for tile_height in range(10, 2, -1):
            grid.tile_height, grid.tile_width = tile_height, tile_height * 2

            if grid.height + 1 < term.height and grid.width <= max_width:
                break
        else:
            with term.location(0, 0):
                print(term.red("terminal size is too small;\n"
                               "please resize the terminal"))
            return False  # game can not continue until after another resize

    margin = (term.width - sum(g.width for g in grids) -
              (len(grids) - 1) * 2) // 2
    for grid_idx, grid in enumerate(grids):
        grid.x = margin + sum(g.width for g in grids[:grid_idx]) + grid_idx * 2

        for row_idx, row in enumerate(grid):
            for col_idx, tile in enumerate(row):
                if tile:
                    tile.x, tile.y = grid.tile_coord(row_idx, col_idx)
                    tile.height, tile.width = grid.tile_height, grid.tile_width

                    grid[row_idx][col_idx] = tile

        grid.draw()
        grid.draw_tiles()

    return True


def main(args=None):
    global do_resize
    do_resize = True

    opts = parser.parse_args(args or sys.argv[1:])

    term = blessed.Terminal()

    DimTile = partial(Tile, base=opts.base, term=term)

    grids = []
    for grid_dim in opts.grid_dims:
        rows, cols = grid_dim

        grid = Grid(x=0, y=1, rows=rows, cols=cols, term=term, Tile=DimTile)
        grid.spawn_tile()
        grid.spawn_tile()

        grids.append(grid)

    def on_resize(signal, frame):
        global do_resize
        do_resize = True
    signal.signal(signal.SIGWINCH, on_resize)

    score = 0
    game_over = False
    term_too_small = False
    with term.fullscreen(), term.cbreak(), term.hidden_cursor():
        while True:
            if do_resize:
                term_too_small = not term_resize(term, grids)
                do_resize = False

            if not term_too_small:
                draw_score(score, term, end=game_over)

            key = term.inkey(_intr_continue=False)
            if key == 'q' or game_over:
                break

            direction = grid_moves.get(key.name or key)
            if not direction or term_too_small:
                continue

            for grid in grids:
                actions = grid.move(direction)

                for action in actions:
                    grid.draw_empty_tile(*action.old)

                    if action.type == Actions.merge:
                        row, column = action.new
                        score += grid[row][column].value

                if actions:  # had any successfull move(s)?
                    grid.spawn_tile(exponent=2 if random.random() > 0.9 else 1)

                    grid.draw_tiles()

                if all(chain(*grid)):
                    for dir in Direction:
                        if grid.move(dir, apply=False):
                            break  # At least one possible move
                    else:  # No possible moves
                        game_over = True

    high = reduce(lambda o, t: max(o, t.value),
                  filter(None, chain(t for g in grids for t in chain(*g))), 0)
    print("highest tile: {}\nscore: {}".format(high, score))

    return 0



import random
from collections import namedtuple
from itertools import product

from enum import Enum

from .tile import Tile

Direction = Enum('Direction', "up left down right")

Position = namedtuple('Position', "row column")

Actions = Enum('Actions', "spawn move merge")


class GridFullError(Exception):
    pass


class GridAction(object):
    def __init__(self, type, new, old=None):
        # Actions:
        self.type = type
        # Position:
        self.new, self.old = new, old

    def __repr__(self):
        return "GridAction({}, new={}{})".format(self.type, self.new,
                ", old=" + str(self.old) if self.old else "")


class Grid(object):
    _grid = [[]]

    def __init__(self, rows=4, cols=4, Tile=Tile):
        self._grid = [[None for _ in range(cols)] for _ in range(rows)]
        self.Tile = Tile

    def __getitem__(self, item):
        return self._grid.__getitem__(item)

    def __setitem__(self, item, value):
        return self._grid.__setitem__(item, value)

    def __delitem__(self, item):
        return self._grid.__delitem__(item)

    def __len__(self):
        return len(self._grid)

    def __repr__(self):
        return "Grid(rows={}, cols={})".format(len(self), len(self[0]))

    def spawn_tile(self, row=None, column=None, apply=True, **kwargs):
        rows, cols = len(self), len(self[0])
        empty_tiles = []
        for row_idx in range(row or 0, row + 1 if row is not None else rows):
            for col_idx in range(
                    column or 0, column + 1 if column is not None else cols):
                if not self[row_idx][col_idx]:
                    empty_tiles.append((row_idx, col_idx))

        if len(empty_tiles) == 0:
            raise GridFullError("no empty tiles")

        row, column = random.choice(empty_tiles)

        if apply:
            self[row][column] = self.Tile(**kwargs)

        return GridAction(Actions.spawn, Position(row, column))

    def move_tile(self, old, new, apply=True):
        old, new = Position(*old), Position(*new)

        if apply:
            self[new.row][new.column] = self[old.row][old.column]
            self[old.row][old.column] = None

        return GridAction(Actions.move, new, old)

    def merge_tiles(self, old, new, apply=True):
        old, new = Position(*old), Position(*new)

        if (not self[new.row][new.column] or
                self[new.row][new.column] != self[old.row][old.column]):
            raise ValueError

        if apply:
            self[new.row][new.column].exponent += 1
            self[old.row][old.column] = None

        return GridAction(Actions.merge, new, old)

    def move_vertical(self, direction, apply=True):
        row_idx_iter = range(len(self))
        if direction is Direction.down:
            row_idx_iter = reversed(row_idx_iter)

        for col_idx, row_idx in product(range(len(self[0])), row_idx_iter):
            if direction is Direction.down:
                pos_in_column = range(row_idx - 1, -1, -1)
            else:
                pos_in_column = range(row_idx + 1, len(self))

            for row_pos in pos_in_column:
                if self[row_idx][col_idx]:  # Tile occupied, maybe merge
                    if self[row_pos][col_idx]:
                        if self[row_idx][col_idx] == self[row_pos][col_idx]:
                            old = Position(row_pos, col_idx)
                            new = Position(row_idx, col_idx)

                            yield self.merge_tiles(old, new, apply=apply)
                        break
                elif self[row_pos][col_idx]:
                    old = Position(row_pos, col_idx)
                    new = Position(row_idx, col_idx)

                    yield self.move_tile(old, new, apply=apply)

                    for row_pos in pos_in_column:
                        if self[row_pos][col_idx]:
                            if self[row_idx][col_idx] == self[row_pos][col_idx]:
                                old = Position(row_pos, col_idx)

                                yield self.merge_tiles(old, new, apply=apply)
                            break
                    break

    def move_horizontal(self, direction, apply=True):
        col_idx_iter = range(len(self[0]))
        if direction is Direction.right:
            col_idx_iter = reversed(col_idx_iter)

        for row_idx, col_idx in product(range(len(self)), col_idx_iter):
            if direction is Direction.right:
                pos_in_row = range(col_idx - 1, -1, -1)
            else:
                pos_in_row = range(col_idx + 1, len(self[0]))

            for col_pos in pos_in_row:
                if self[row_idx][col_idx]:
                    if self[row_idx][col_pos]:
                        if self[row_idx][col_idx] == self[row_idx][col_pos]:
                            old = Position(row_idx, col_pos)
                            new = Position(row_idx, col_idx)

                            yield self.merge_tiles(old, new, apply=apply)
                        break
                elif self[row_idx][col_pos]:
                    old = Position(row_idx, col_pos)
                    new = Position(row_idx, col_idx)

                    yield self.move_tile(old, new, apply=apply)

                    for col_pos in pos_in_row:
                        if self[row_idx][col_pos]:
                            if self[row_idx][col_idx] == self[row_idx][col_pos]:
                                old = Position(row_idx, col_pos)

                                yield self.merge_tiles(old, new, apply=apply)
                            break
                    break

    def move(self, direction, apply=True):
        if not isinstance(direction, Direction):
            raise TypeError

        if direction in (Direction.up, Direction.down):
            return list(self.move_vertical(direction, apply))
        return list(self.move_horizontal(direction, apply))

    def resize(self, rows=None, cols=None):
        if rows:
            if rows < len(self):
                del self[rows:]
            elif rows > len(self):
                columns = cols or len(self[0])
                for _ in range(rows - len(self)):
                    self._grid.append([None for _ in range(columns)])

        if cols:
            for row in self:
                if cols < len(row):
                    del row[cols:]
                elif cols > len(row):
                    row.extend([None for _ in range(cols - len(row))])



class Tile(object):
    _base = 0
    _exponent = 1

    def __init__(self, base=2, exponent=1):
        self.base = base
        self.exponent = exponent

    @property
    def base(self):
        return self._base

    @base.setter
    def base(self, value):
        self._base = value
        self.value = self._base ** self.exponent

    @property
    def exponent(self):
        return self._exponent

    @exponent.setter
    def exponent(self, value):
        self._exponent = value
        self.value = self.base ** self._exponent

    def __eq__(self, other):
        return (isinstance(other, Tile) and self.base == other.base and
                self.exponent == other.exponent)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, Tile):
            raise NotImplementedError

        return self.value < other.value

    def __repr__(self):
        return "Tile(base={}, exponent={})".format(self.base, self.exponent)



from collections import namedtuple

components = 'major, minor, micro, releaselevel, serial'
values = (0, 1, 2, 'final', 0)
__version_info__ = namedtuple('__version_info__', components)(*values)
__version__ = (str(__version_info__.major) + '.' +
               str(__version_info__.minor) + '.' +
               str(__version_info__.micro))
del components, values



from .. import tile


class Tile(tile.Tile):
    def __init__(self, base=2, exponent=1, x=0, y=0, width=0, height=0,
                 term=None):
        super(Tile, self).__init__(base=base, exponent=exponent)
        self.x, self.y = x, y
        self.width, self.height = width, height

        self.term = term

    def draw(self, text=None, fg='white', bg='black'):
        text = text or str(self.value)
        style = getattr(self.term, fg + ("_on_" + bg if bg else ""))

        for y_offset in range(self.height):
            with self.term.location(self.x, self.y + y_offset):
                print(style(' ' * self.width))

        hor_offset = (self.width + 1) // 2 - (len(text) + 1) // 2
        vert_offset = (self.height + 1) // 2 - 1
        with self.term.location(self.x + hor_offset, self.y + vert_offset):
            print(style(text))



#!/usr/bin/env python

import os
from setuptools import setup, find_packages

import macht


setup(
    name='macht',
    version=macht.__version__,
    description="A 2048 clone in python with Terminal UI",
    long_description=open(os.path.join(os.path.dirname(__file__),
                          'README.rst')).read(),
    author="Rolf Morel",
    author_email="rolfmorel@gmail.com",
    url="https://github.com/polyphemus/macht",
    packages=find_packages(exclude=["tests"]),
    license='LGPLv3',
    install_requires=['blessed', 'enum34'],
    entry_points={'console_scripts': ('macht = macht.term:main')},
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: '
        'GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Topic :: Games/Entertainment :: Puzzle Games',
        'Topic :: Terminals'
        ],
    keywords=['2048', 'terminal', 'game', 'board', 'puzzle', 'tiles']
)



from itertools import chain

from .. import grid
from . import tile

bg_colors = ['red', 'magenta', 'green', 'cyan', 'blue', 'cyan', 'yellow']
fg_colors = ['black', 'black', 'black', 'black', 'white', 'black', 'black']


class Grid(grid.Grid):
    vert_div = '||'
    hor_div = '='
    cross_div = 'XX'

    def __init__(self, x=0, y=0, rows=0, cols=0, tile_width=0, tile_height=0,
                 term=None, Tile=tile.Tile):
        super(Grid, self).__init__(rows=rows, cols=cols, Tile=Tile)
        self.x, self.y = x, y
        self.tile_width, self.tile_height = tile_width, tile_height
        self.term = term

    def draw(self, fg='white', bg=None):
        style = getattr(self.term, fg + ("_on_" + bg if bg else ""))
        rows, cols = len(self), len(self[0])

        for col_idx in range(cols - 1):
            hor_offset = ((col_idx + 1) * self.tile_width +
                          col_idx * len(self.vert_div))

            for vert_offset in range(self.height):
                with self.term.location(self.x + hor_offset,
                                        self.y + vert_offset):
                    print(style(self.vert_div))

        for row_idx in range(rows - 1):
            vert_offset = (row_idx + 1) * self.tile_height + row_idx
            with self.term.location(self.x, self.y + vert_offset):
                print(style(self.cross_div.join(
                      [self.hor_div * self.tile_width] * cols)))

    def draw_tiles(self):
        for tile in filter(None, chain(*self)):  # all non-empty tiles
            # choose a color, use modulo to support any value of a tile
            # all bg colors have a bright variant doubling the amount of colors
            color_idx = (tile.exponent - 1) % (len(bg_colors) * 2) // 2
            bg = bg_colors[color_idx]
            if self.term.number_of_colors >= 16 and tile.exponent % 2 == 0:
                bg = "bright_" + bg

            tile.draw(fg=fg_colors[color_idx], bg=bg)

    def draw_empty_tile(self, row, column):
        x, y = self.tile_coord(row, column)
        for y_offset in range(self.tile_height):
            with self.term.location(x, y + y_offset):
                print(' ' * self.tile_width)

    @property
    def width(self):
        cols = len(self[0])
        if cols > 0:
            return cols * self.tile_width + (cols - 1) * len(self.vert_div)
        return 0

    @property
    def height(self):
        rows = len(self)
        if rows > 0:
            return rows * self.tile_height + rows - 1
        return 0

    def tile_coord(self, row, column):
        if row >= len(self) or column >= len(self[0]):
            raise IndexError

        x = self.x + column * self.tile_width + column * len(self.vert_div)
        y = self.y + row * self.tile_height + row

        return x, y

    def spawn_tile(self, *args, **kwargs):
        action = super(Grid, self).spawn_tile(*args, **kwargs)
        row, column = action.new

        self[row][column].x, self[row][column].y = self.tile_coord(row, column)
        self[row][column].height = self.tile_height
        self[row][column].width = self.tile_width

        return action

    def move(self, *args, **kwargs):
        actions = super(Grid, self).move(*args, **kwargs)

        for action in actions:
            row, col = action.new

            self[row][col].x, self[row][col].y = self.tile_coord(row, col)

        return actions
